import java.util.*;

public class Problem {

    String name;
    ArrayList<IntVar> variables;
    ArrayList<Constraint> constraints;
    Stack<Constraint> revisionStack;
    Stack<ArrayList<ReversibleVar>> trail;
    boolean trace;
    int world, solutions, nodes, fails;
    IntVar[] var;
    int[] val;
    boolean firstProbe;

    public Problem(String name){
	this.name     = name;
	variables     = new ArrayList<IntVar>();
	constraints   = new ArrayList<Constraint>();
	revisionStack = new Stack<Constraint>();
	trail         = new Stack<ArrayList<ReversibleVar>>();
	trace         = false;
	world         = 0;
	solutions     = 0;
	firstProbe    = true;
	nodes         = 0;
	fails         = 0;
    }

    IntVar intVar(String name,int lwb,int upb){
	IntVar v = new IntVar(name,lwb,upb,this);
	return v;
    }

    Constraint post(Constraint c){
	constraints.add(c);
	revisionStack.push(c);
	c.onStack = true;
	c.pb = this;
	return c;
    }

    ReversibleInt reversibleInt(){
	return new ReversibleInt(this,Integer.MIN_VALUE);
    }

    ReversibleBool reversibleBool(){
	return new ReversibleBool(this,false);
    }

    boolean propagate(){
	boolean consistent = true;
	while (!revisionStack.isEmpty() && consistent){
	    Constraint c = revisionStack.pop();
	    if (c.init){
		try {c.propagateInit();}
		catch (CPException e) {consistent = false;}
		c.init = false;
	    }
	    while (consistent && !c.removals.isEmpty()){
		IntVar v = c.removals.remove();
		while (consistent && !v.deletions.isEmpty()){
		    int x = v.deletions.remove();
		    try {c.propagateRemoval(v,x);}
		    catch (CPException e) {consistent = false;}
		}
	    }
	    while (consistent && !c.newUPB.isEmpty()){
		IntVar v = c.newUPB.remove();
		try {c.propagateUPB(v);}
		catch (CPException e) {consistent = false;}
	    }
	    while (consistent && !c.newLWB.isEmpty()){
		IntVar v = c.newLWB.remove();
		try {c.propagateLWB(v);}
		catch (CPException e) {consistent = false;}
	    }
	    while (consistent && !c.instantiations.isEmpty()){
		IntVar v = c.instantiations.remove();
		try {c.propagateInstantiate(v);}
		catch (CPException e) {consistent = false;}
	    }
	    c.onStack = false;
	    assert c.isLegal() : c;
	}
	if (!consistent) flushRevisionStack();
	return consistent;
    }

    void flushRevisionStack(){
	while (!revisionStack.isEmpty()){
	    Constraint c = revisionStack.pop();
	    while (!c.removals.isEmpty()){
		IntVar v = c.removals.remove();
		v.deletions.clear();
	    }
	    c.newUPB.clear();
	    c.newLWB.clear();
	    c.instantiations.clear();
	    c.onStack = false;
	}
    }

    void pushWorld(){
	trail.push(new ArrayList<ReversibleVar>());
	world++;
    }

    void popWorld(){
	ArrayList<ReversibleVar> current = trail.pop();
	for (ReversibleVar v : current) v.restoreValue();
	world--;
    }

    void reset(){
	while (world > 0) popWorld();
    }	

    public boolean bt() {
	nodes = 0; fails = 0;
	try {if (propagate()) bt0();}
	catch (SolutionException e) {return true;}
	return false;
    }

    private void bt0() {
	IntVar v = selectVar();
	if (v == null) throw new SolutionException();
	for (int x : v){
	    nodes++;
	    pushWorld();
	    v.instantiate(x);
	    if (propagate()) bt0(); else fails++;
	    popWorld();
	}
    }  

    public int btCount() {
	solutions = 0;
	if (propagate()) btCount0();
	return solutions;
    }
    
    private void btCount0() {
	IntVar v = selectVar();
	if (v == null){solutions++; return;}
	for (int x : v){
	    pushWorld();
	    v.instantiate(x);
	    if (propagate()) btCount0();
	    popWorld();
	}
    }

    public boolean btProbe(){
	if (firstProbe){
	    int n = variables.size();
	    var   = new IntVar[n+1];
	    val   = new int[n+1];
	    Arrays.fill(var,null);
	    Arrays.fill(val,Integer.MIN_VALUE);
	    firstProbe = false;
	}
	reset();
	try {if (propagate()) btProbe0();}
	catch (SolutionException e) {return true;}
	return false;
    }
    
    public void btProbe0() {
	IntVar v = null;
	if (var[world] == null)
	    v = selectVar();
	else
	    v = var[world];
	if (v == null){
	    val[world - 1]++;
	    throw new SolutionException();
	}
	var[world] = v;
	for (int x : v){
	    if (x >= val[world]){
		val[world] = x;
		pushWorld();
		v.instantiate(x);
		if (propagate()) btProbe0();
		popWorld();
	    }
	}
	var[world] = null;
	val[world] = Integer.MIN_VALUE;
    }

   
    public IntVar selectVar(){
	for (IntVar v : variables)
	    if (!v.isInstantiated()) return v;
	return null;
    }

    void show(){
	System.out.println(name);
	for (IntVar v : variables) System.out.print(v +" ");
	System.out.println();
	for (Constraint c : constraints) System.out.println(c);
    }

    void showBrief(){
	for (IntVar v : variables) System.out.print(v +" ");
	System.out.println();
    }

    void showRevisionStack(){
	for (Constraint c : revisionStack) System.out.println(c);
    }
}
