package weeSeepy.problem;

import java.util.*;
import weeSeepy.variables.*;
import weeSeepy.constraints.*;
import weeSeepy.reversibles.*;
import weeSeepy.heuristics.*;
import weeSeepy.exceptions.*;;

public class Problem {

    String name;
    public ArrayList<IntVar> variables;
    ArrayList<Constraint> constraints;
    public LinkedList<Constraint> revisionQueue;
    public Stack<ArrayList<ReversibleVar>> trail;
    public VarOrdHeur voh;
    public boolean trace;
    public int world, solutions, nodes, fails;
    public long cpuTime;
    IntVar[] var;
    int[] val;
    public boolean firstProbe, propagationOn;

    public Problem(String name){
	this.name     = name;
	variables     = new ArrayList<IntVar>();
	constraints   = new ArrayList<Constraint>();
	revisionQueue = new LinkedList<Constraint>();
	trail         = new Stack<ArrayList<ReversibleVar>>();
	voh           = new IndexOrder(variables);
	trace         = false;
	world         = 0;
	solutions     = 0;
	firstProbe    = true;
	propagationOn = true;
	nodes         = 0;
	fails         = 0;
    }

    public IntVar intVar(String name,int lwb,int upb){
	IntVar v = new IntVar(name,lwb,upb,this);
	return v;
    }

    public Constraint post(Constraint c){
	if (!c.posted){
	    constraints.add(c);
	    revisionQueue.add(c);
	    for (IntVar x: c.variables)
		x.constraints.add(c);
	    c.onQueue = true;
	    c.pb = this;
	    c.posted = true;
	}
	return c;
    }

    public ReversibleInt reversibleInt(){
	return new ReversibleInt(this,Integer.MIN_VALUE);
    }

    public ReversibleBool reversibleBool(){
	return new ReversibleBool(this,false);
    }

    public boolean propagate(){
	boolean consistent = true;
	while (consistent && !revisionQueue.isEmpty()){
	    Constraint c = revisionQueue.peek();
	    consistent = c.revise();
	    if (consistent) {
		revisionQueue.remove();
		c.onQueue = false;
		assert c.isLegal() : c;
	    }
	    if (!consistent) flushRevisionQueue();
	}
	assert cleanConstraints() : "End of propagate and constraint queue is not clean";
	return consistent;
    }
    
    void flushRevisionQueue(){
	while (!revisionQueue.isEmpty()){
	    Constraint c = revisionQueue.remove();
	    if (trace) System.out.println("FLUSH: "+ c);
	    while (!c.removals.isEmpty()){
		IntVar v = c.removals.remove();
		v.deletions.clear();
	    }
	    c.newUPB.clear();
	    c.newLWB.clear();
	    c.instantiations.clear();
	    c.onQueue = false;
	}
    }

    public void pushWorld(){
	trail.push(new ArrayList<ReversibleVar>());
	world++;
    }

    public void popWorld(){
	ArrayList<ReversibleVar> current = trail.pop();
	for (ReversibleVar v : current) v.restoreValue();
	world--;
    }

    public void reset(){
	while (world > 0) popWorld();
    }

    public boolean solve() {
	if (trace) show();
	nodes = 0; fails = 0; cpuTime = System.currentTimeMillis();
	boolean solved = false;
	try {if (propagate()) dfs();}
	catch (SolutionException e){solved = true;}
	cpuTime = System.currentTimeMillis() - cpuTime;
	return solved;
    }

    private void dfs() {
	IntVar v = voh.select();
	if (trace) System.out.println("dfs: selected "+ v);
	if (v == null) throw new SolutionException();
	for (int x : v){
	    nodes++;
	    pushWorld();
	    if (trace) showBrief();
	    v.instantiate(x);
	    if (propagate())
		dfs();
	    else fails++;
	    popWorld();
	}
    }  

    public int count() {
	nodes = 0; fails = 0; cpuTime = System.currentTimeMillis();
	solutions = 0;
	if (propagate()) enumerate();
	cpuTime = System.currentTimeMillis() - cpuTime;
	return solutions;
    }
    
    private void enumerate() {
	IntVar v = voh.select();
	if (v == null){solutions++; return;}
	for (int x : v){
	    pushWorld();
	    v.instantiate(x);
	    if (propagate()) enumerate();
	    popWorld();
	}
    }

    public boolean nextSolution(){
	if (firstProbe){
	    nodes = 0; fails = 0; cpuTime = System.currentTimeMillis();
	    int n = variables.size();
	    var   = new IntVar[n+1];
	    val   = new int[n+1];
	    Arrays.fill(var,null);
	    Arrays.fill(val,Integer.MIN_VALUE);
	    firstProbe = false;
	}
	reset();
	boolean solved = false;
	try {if (propagate()) probe();}
	catch (SolutionException e) {solved = true;}
	cpuTime = System.currentTimeMillis() - cpuTime;
	return solved;
    }
    
    private void probe() {
	IntVar v = null;
	if (var[world] == null)
	    v = voh.select();
	else
	    v = var[world];
	if (v == null){
	    val[world - 1]++;
	    throw new SolutionException();
	}
	var[world] = v;
	for (int x : v){
	    if (x >= val[world]){
		val[world] = x;
		pushWorld();
		v.instantiate(x);
		if (propagate()) probe();
		popWorld();
	    }
	}
	var[world] = null;
	val[world] = Integer.MIN_VALUE;
    }

   
    public IntVar selectVar(){
	for (IntVar v : variables)
	    if (!v.isInstantiated()) return v;
	return null;
    }

    public void show(){
	System.out.println(name +" world: "+ world);
	for (IntVar v : variables) System.out.print(v +" ");
	System.out.println();
	for (Constraint c : constraints) System.out.println(c);
    }

    public void showBrief(){
	System.out.println(name +" world: "+ world);
	for (IntVar v : variables) System.out.print(v +" ");
	System.out.println();
    }

    public void showRevisionQueue(){
	System.out.print("REVISION Queue size: "+ revisionQueue.size() +" ");
	for (Constraint c : revisionQueue) System.out.println(c);
    }

    boolean cleanConstraints(){
	for (Constraint c : constraints)
	    if (c.onQueue || c.removals.size() > 0 || c.newUPB.size() > 0 || c.newLWB.size() > 0 || c.instantiations.size() > 0)
		return false;
	return true;
    }
}
